# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import pathlib
from typing import Generator

import numpy as np
import pytest
import scipp as sc
import scippnexus as snx

from ess.nmx.data import small_mcstas_2_sample, small_mcstas_3_sample
from ess.nmx.mcstas_loader import (
    DefaultMaximumProbability,
    InputFilepath,
    event_weights_from_probability,
    load_mcstas_nexus,
    proton_charge_from_event_data,
)
from ess.nmx.reduction import NMXData


def check_scalar_properties_mcstas_2(dg: NMXData):
    """Test helper for NMXData loaded from McStas 2.

    Expected numbers are hard-coded based on the sample file.
    """

    assert dg.proton_charge == sc.scalar(0.15430000000000002, unit=None)
    assert sc.identical(dg.crystal_rotation, sc.vector([20, 0, 90], unit='deg'))
    assert sc.identical(dg.sample_position, sc.vector(value=[0, 0, 0], unit='m'))
    assert sc.identical(
        dg.source_position, sc.vector(value=[-0.53123, 0.0, -157.405], unit='m')
    )
    assert dg.sample_name == sc.scalar("sampleMantid")


def check_nmxdata_properties(dg: NMXData) -> None:
    assert isinstance(dg, sc.DataGroup)
    assert dg.shape == (3, 1280 * 1280)
    # Check maximum value of weights.
    assert sc.identical(
        dg.weights.max().data,
        sc.scalar(DefaultMaximumProbability, unit='counts', dtype=float),
    )
    # Expected values are provided by the IDS
    # based on the simulation settings of the sample file.
    assert np.all(
        np.round(dg.fast_axis.values, 2)
        == sc.vectors(
            dims=['panel'],
            values=[(1.0, 0.0, -0.01), (-0.01, 0.0, -1.0), (0.01, 0.0, 1.0)],
        ).values,
    )
    assert sc.identical(
        dg.slow_axis,
        sc.vectors(
            dims=['panel'], values=[[0.0, 1.0, 0.0], [0.0, 1.0, 0.0], [0.0, 1.0, 0.0]]
        ),
    )


def test_file_reader_mcstas2(mcstas_2_deprecation_warning_context) -> None:
    with mcstas_2_deprecation_warning_context():
        file_path = InputFilepath(small_mcstas_2_sample())

    entry_path = "entry1/data/bank01_events_dat_list_p_x_y_n_id_t"
    with snx.File(file_path) as file:
        raw_data = file[entry_path]["events"][()]
        data_length = raw_data.sizes['dim_0']

    dg = load_mcstas_nexus(
        file_path=file_path,
        event_weights_converter=event_weights_from_probability,
        proton_charge_converter=proton_charge_from_event_data,
    )
    check_scalar_properties_mcstas_2(dg)
    assert dg.weights.bins.size().sum().value == data_length
    check_nmxdata_properties(dg)


def check_scalar_properties_mcstas_3(dg: NMXData):
    """Test helper for NMXData loaded from McStas 3.

    Expected numbers are hard-coded based on the sample file.
    """

    assert dg.proton_charge == sc.scalar(0.0167, unit=None)
    assert sc.identical(dg.crystal_rotation, sc.vector([0, 0, 0], unit='deg'))
    assert sc.identical(dg.sample_position, sc.vector(value=[0, 0, 0], unit='m'))
    assert sc.identical(
        dg.source_position, sc.vector(value=[-0.53123, 0.0, -157.405], unit='m')
    )
    assert dg.sample_name == sc.scalar("sampleMantid")


def test_file_reader_mcstas3() -> None:
    file_path = InputFilepath(small_mcstas_3_sample())
    entry_paths = [
        f"entry1/data/bank0{i}_events_dat_list_p_x_y_n_id_t" for i in range(1, 4)
    ]
    with snx.File(file_path) as file:
        raw_datas = [file[entry_path]["events"][()] for entry_path in entry_paths]
        raw_data = sc.concat(raw_datas, dim='dim_0')
        data_length = raw_data.sizes['dim_0']

    dg = load_mcstas_nexus(
        file_path=file_path,
        event_weights_converter=event_weights_from_probability,
        proton_charge_converter=proton_charge_from_event_data,
    )
    check_scalar_properties_mcstas_3(dg)
    assert dg.weights.bins.size().sum().value == data_length
    check_nmxdata_properties(dg)


@pytest.fixture(params=[small_mcstas_2_sample, small_mcstas_3_sample])
def tmp_mcstas_file(
    tmp_path: pathlib.Path,
    request: pytest.FixtureRequest,
    mcstas_2_deprecation_warning_context,
) -> Generator[pathlib.Path, None, None]:
    import os
    import shutil

    if request.param == small_mcstas_2_sample:
        with mcstas_2_deprecation_warning_context():
            original_file_path = request.param()
    else:
        original_file_path = request.param()

    tmp_file = tmp_path / pathlib.Path('file.h5')
    shutil.copy(original_file_path, tmp_file)
    yield tmp_file
    os.remove(tmp_file)


def test_file_reader_mcstas_additional_fields(tmp_mcstas_file: pathlib.Path) -> None:
    """Check if additional fields names do not break the loader."""
    import h5py

    entry_path = "entry1/data/bank01_events_dat_list_p_x_y_n_id_t"
    new_entry_path = entry_path + '_L'

    with h5py.File(tmp_mcstas_file, 'r+') as file:
        dataset = file[entry_path]
        del file[entry_path]
        file[new_entry_path] = dataset

    dg = load_mcstas_nexus(
        file_path=InputFilepath(str(tmp_mcstas_file)),
        event_weights_converter=event_weights_from_probability,
        proton_charge_converter=proton_charge_from_event_data,
    )

    assert isinstance(dg, sc.DataGroup)
    assert dg.shape == (3, 1280 * 1280)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

_version = '0'

__all__ = ['small_mcstas_2_sample', 'small_mcstas_3_sample', 'get_path']


def _make_pooch():
    import pooch

    return pooch.create(
        path=pooch.os_cache('essnmx'),
        env='ESSNMX_DATA_DIR',
        retry_if_failed=3,
        base_url='https://public.esss.dk/groups/scipp/ess/nmx/',
        version=_version,
        registry={
            'small_mcstas_2_sample.h5': 'md5:c3affe636397f8c9eea1d9c10a2bf487',
            'small_mcstas_3_sample.h5': 'md5:2afaac205d13ee857ee5364e3f1957a7',
        },
    )


_pooch = _make_pooch()


def small_mcstas_2_sample():
    """McStas 2 file containing small number of events."""
    import warnings

    warnings.warn(
        DeprecationWarning(
            '``essnmx`` will not support loading files '
            'made by McStas with version less than 3 from ``25.0.0``. '
            'Use ``small_mcstas_3_sample`` instead.'
        ),
        stacklevel=2,
    )

    return get_path('small_mcstas_2_sample.h5')


def small_mcstas_3_sample():
    """McStas 3 file that contains only ``bank0(1-3)`` in the ``data`` group.

    Real McStas 3 file should contain more dataset under ``data`` group.
    """
    return get_path('small_mcstas_3_sample.h5')


def get_path(name: str) -> str:
    """
    Return the path to a data file bundled with ess nmx.

    This function only works with example data and cannot handle
    paths to custom files.
    """
    return _pooch.fetch(name)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import io
import pathlib
from typing import NewType, Optional, Union, overload

import h5py
import scipp as sc

TimeBinSteps = NewType("TimeBinSteps", int)


class _SharedFields(sc.DataGroup):
    """All shared fields between NMXData and NMXReducedData.

    ``weights`` is only present in NMXData due to potential memory issues.
    """

    @property
    def origin_position(self) -> sc.Variable:
        """Position of the first pixel (lowest ID) in the detector.

        Relative position from the sample."""
        return self['origin_position']

    @property
    def crystal_rotation(self) -> sc.Variable:
        """Rotation of the crystal."""

        return self['crystal_rotation']

    @property
    def sample_name(self) -> sc.Variable:
        return self['sample_name']

    @property
    def fast_axis(self) -> sc.Variable:
        """Fast axis, along where the pixel ID increases by 1."""
        return self['fast_axis']

    @property
    def slow_axis(self) -> sc.Variable:
        """Slow axis, along where the pixel ID increases by > 1.

        The pixel ID increases by the number of pixels in the fast axis."""
        return self['slow_axis']

    @property
    def proton_charge(self) -> sc.Variable:
        """Accumulated number of protons during the measurement."""
        return self['proton_charge']

    @property
    def source_position(self) -> sc.Variable:
        """Relative position of the source from the sample."""
        return self['source_position']

    @property
    def sample_position(self) -> sc.Variable:
        """Relative position of the sample from the sample. (0, 0, 0)"""
        return self['sample_position']


class NMXData(_SharedFields, sc.DataGroup):
    @property
    def weights(self) -> sc.DataArray:
        """Event data grouped by pixel id and panel."""
        return self['weights']


class NMXReducedData(_SharedFields, sc.DataGroup):
    @property
    def counts(self) -> sc.DataArray:
        """Binned time of arrival data from flattened event data."""
        return self['counts']

    def _create_dataset_from_var(
        self,
        *,
        root_entry: h5py.Group,
        var: sc.Variable,
        name: str,
        long_name: Optional[str] = None,
        compression: Optional[str] = None,
        compression_opts: Optional[int] = None,
    ) -> h5py.Dataset:
        compression_options = dict()
        if compression is not None:
            compression_options["compression"] = compression
        if compression_opts is not None:
            compression_options["compression_opts"] = compression_opts

        dataset = root_entry.create_dataset(
            name,
            data=var.values,
            **compression_options,
        )
        dataset.attrs["units"] = str(var.unit)
        if long_name is not None:
            dataset.attrs["long_name"] = long_name
        return dataset

    def _create_compressed_dataset(
        self,
        *,
        root_entry: h5py.Group,
        name: str,
        var: sc.Variable,
        long_name: Optional[str] = None,
    ) -> h5py.Dataset:
        return self._create_dataset_from_var(
            root_entry=root_entry,
            var=var,
            name=name,
            long_name=long_name,
            compression="gzip",
            compression_opts=4,
        )

    def _create_root_data_entry(self, file_obj: h5py.File) -> h5py.Group:
        nx_entry = file_obj.create_group("NMX_data")
        nx_entry.attrs["NX_class"] = "NXentry"
        nx_entry.attrs["default"] = "data"
        nx_entry.attrs["name"] = "NMX"
        nx_entry["name"] = "NMX"
        nx_entry["definition"] = "TOFRAW"
        return nx_entry

    def _create_sample_group(self, nx_entry: h5py.Group) -> h5py.Group:
        nx_sample = nx_entry.create_group("NXsample")
        nx_sample["name"] = self.sample_name.value
        # Crystal rotation
        self._create_dataset_from_var(
            root_entry=nx_sample,
            var=self.crystal_rotation,
            name='crystal_rotation',
            long_name='crystal rotation in Phi (XYZ)',
        )
        return nx_sample

    def _create_instrument_group(self, nx_entry: h5py.Group) -> h5py.Group:
        nx_instrument = nx_entry.create_group("NXinstrument")
        nx_instrument.attrs["nr_detector"] = self.origin_position.sizes['panel']
        nx_instrument.create_dataset("proton_charge", data=self.proton_charge.value)

        nx_detector_1 = nx_instrument.create_group("detector_1")
        # Detector counts
        self._create_compressed_dataset(
            root_entry=nx_detector_1,
            name="counts",
            var=self.counts.fold(
                'id', sizes={'panel': 1, 'id': self.counts.sizes['id']}
            ),
        )
        # Time of arrival bin edges
        self._create_dataset_from_var(
            root_entry=nx_detector_1,
            var=self.counts.coords['t'],
            name="t_bin",
            long_name="t_bin TOF (ms)",
        )
        # Pixel IDs
        self._create_compressed_dataset(
            root_entry=nx_detector_1,
            name="pixel_id",
            var=self.counts.coords['id'],
            long_name="pixel ID",
        )
        return nx_instrument

    def _create_detector_group(self, nx_entry: h5py.Group) -> h5py.Group:
        nx_detector = nx_entry.create_group("NXdetector")
        # Position of the first pixel (lowest ID) in the detector
        self._create_compressed_dataset(
            root_entry=nx_detector,
            name="origin",
            var=self.origin_position,
        )
        # Fast axis, along where the pixel ID increases by 1
        self._create_dataset_from_var(
            root_entry=nx_detector, var=self.fast_axis, name="fast_axis"
        )
        # Slow axis, along where the pixel ID increases
        # by the number of pixels in the fast axis
        self._create_dataset_from_var(
            root_entry=nx_detector, var=self.slow_axis, name="slow_axis"
        )
        return nx_detector

    def _create_source_group(self, nx_entry: h5py.Group) -> h5py.Group:
        nx_source = nx_entry.create_group("NXsource")
        nx_source["name"] = "European Spallation Source"
        nx_source["short_name"] = "ESS"
        nx_source["type"] = "Spallation Neutron Source"
        nx_source["distance"] = sc.norm(self.source_position).value
        nx_source["probe"] = "neutron"
        nx_source["target_material"] = "W"
        return nx_source

    @overload
    def export_as_nexus(self, output_file_base: str) -> None:
        ...

    @overload
    def export_as_nexus(self, output_file_base: pathlib.Path) -> None:
        ...

    @overload
    def export_as_nexus(self, output_file_base: io.BytesIO) -> None:
        ...

    def export_as_nexus(
        self, output_file_base: Union[str, pathlib.Path, io.BytesIO]
    ) -> None:
        """Export the reduced data to a NeXus file.

        Currently exporting step is not expected to be part of sciline pipelines.
        """
        if isinstance(output_file_base, (str, pathlib.Path)):
            file_base = pathlib.Path(output_file_base)
            if file_base.suffix not in (".h5", ".nxs"):
                raise ValueError("Output file name must end with .h5 or .nxs")
        else:
            file_base = output_file_base

        with h5py.File(file_base, "w") as out_file:
            out_file.attrs["default"] = "NMX_data"
            # Root Data Entry
            nx_entry = self._create_root_data_entry(out_file)
            # Sample
            self._create_sample_group(nx_entry)
            # Instrument
            self._create_instrument_group(nx_entry)
            # Detector
            self._create_detector_group(nx_entry)
            # Source
            self._create_source_group(nx_entry)


def bin_time_of_arrival(
    nmx_data: NMXData, time_bin_step: TimeBinSteps
) -> NMXReducedData:
    """Bin time of arrival data into ``time_bin_step`` bins."""

    counts: sc.DataArray = nmx_data.weights.flatten(dims=['panel', 'id'], to='id').hist(
        t=time_bin_step
    )
    counts.unit = 'counts'

    return NMXReducedData(
        counts=counts,
        **{key: nmx_data[key] for key in nmx_data.keys() if key != 'weights'},
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

# flake8: noqa
import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

del importlib

from .data import small_mcstas_3_sample
from .mcstas_loader import InputFilepath, NMXData, load_mcstas_nexus
from .reduction import NMXReducedData, TimeBinSteps

__all__ = [
    "small_mcstas_3_sample",
    "NMXData",
    "InputFilepath",
    "load_mcstas_nexus",
    "NMXReducedData",
    "TimeBinSteps",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import io

import pytest

from ess.nmx.reduction import NMXReducedData


@pytest.fixture
def reduced_data() -> NMXReducedData:
    import numpy as np
    import scipp as sc

    rng = np.random.default_rng(42)
    id_list = sc.array(dims=['event'], values=rng.integers(0, 12, size=100))
    t_list = sc.array(dims=['event'], values=rng.random(size=100, dtype=float))
    counts = (
        sc.DataArray(
            data=sc.ones(dims=['event'], shape=[100]),
            coords={'id': id_list, 't': t_list},
        )
        .group('id')
        .hist(t=10)
    )

    return NMXReducedData(
        counts=counts,
        proton_charge=sc.scalar(1.0, unit='counts'),
        crystal_rotation=sc.vector(value=[0.0, 20.0, 0.0], unit='deg'),
        fast_axis=sc.vectors(
            dims=['panel'],
            values=[[1.0, 0.0, 0.0], [1.0, 0.0, 0.0], [1.0, 0.0, 0.0]],
            unit='m',
        ),
        slow_axis=sc.vectors(
            dims=['panel'],
            values=[[0.0, 1.0, 0.0], [0.0, 1.0, 0.0], [0.0, 1.0, 0.0]],
            unit='m',
        ),
        origin_position=sc.vectors(
            dims=['panel'],
            values=[[-0.2, 0.0, 0.0], [0.0, 0.0, 0.0], [0.2, 0.0, 0.0]],
            unit='m',
        ),
        sample_position=sc.vector(value=[0.0, 0.0, 0.0], unit='m'),
        source_position=sc.vector(value=[-3, 0.0, -4], unit='m'),
        sample_name=sc.scalar('Unit Test Sample'),
        position=sc.zeros(dims=['panel', 'id'], shape=[3, 4], unit='m'),
    )


def test_mcstas_reduction_export_to_bytestream(reduced_data: NMXReducedData) -> None:
    """Test export method."""
    import h5py
    import numpy as np
    import scipp as sc

    data_fields = [
        'NXdetector',
        'NXsample',
        'NXsource',
        'NXinstrument',
        'definition',
        'name',
    ]

    with io.BytesIO() as bio:
        reduced_data.export_as_nexus(bio)
        with h5py.File(bio, 'r') as f:
            assert 'NMX_data' in f
            nmx_data: h5py.Group = f.require_group('NMX_data')
            for field in data_fields:
                assert field in nmx_data

            nx_detector = nmx_data.require_group('NXdetector')
            assert np.all(nx_detector['fast_axis'][()] == reduced_data.fast_axis.values)
            assert np.all(nx_detector['slow_axis'][()] == reduced_data.slow_axis.values)
            assert np.all(
                nx_detector['origin'][()] == reduced_data.origin_position.values
            )

            instrument_data = nmx_data.require_group('NXinstrument')
            assert (
                instrument_data['proton_charge'][()] == reduced_data.proton_charge.value
            )

            det1_data = instrument_data.require_group('detector_1')
            assert np.all(det1_data['counts'][()] == reduced_data.counts.values)
            assert np.all(
                det1_data['pixel_id'][()] == reduced_data.counts.coords['id'].values
            )
            assert np.all(
                det1_data['t_bin'][()] == reduced_data.counts.coords['t'].values
            )

            nx_sample = nmx_data.require_group('NXsample')
            sample_name: bytes = nx_sample['name'][()]
            assert sample_name.decode() == reduced_data.sample_name.value

            nx_source = nmx_data.require_group('NXsource')
            assert (
                nx_source['distance'][()] == sc.norm(reduced_data.source_position).value
            )


def test_mcstas_reduction_export_wrong_extension_raises(
    reduced_data: NMXReducedData,
) -> None:
    """Test export method."""
    with pytest.raises(ValueError, match="Output file name must end with .h5 or .nxs"):
        reduced_data.export_as_nexus('wrong_name.txt')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import pytest
import sciline as sl
import scipp as sc

from ess.nmx.data import small_mcstas_2_sample, small_mcstas_3_sample


@pytest.fixture(params=[small_mcstas_2_sample, small_mcstas_3_sample])
def mcstas_file_path(
    request: pytest.FixtureRequest, mcstas_2_deprecation_warning_context
) -> str:
    if request.param == small_mcstas_2_sample:
        with mcstas_2_deprecation_warning_context():
            return request.param()

    return request.param()


@pytest.fixture
def mcstas_workflow(mcstas_file_path: str) -> sl.Pipeline:
    from ess.nmx.mcstas_loader import (
        DefaultMaximumProbability,
        EventWeightsConverter,
        InputFilepath,
        MaximumProbability,
        ProtonChargeConverter,
        event_weights_from_probability,
        load_mcstas_nexus,
        proton_charge_from_event_data,
    )
    from ess.nmx.reduction import TimeBinSteps, bin_time_of_arrival

    return sl.Pipeline(
        [load_mcstas_nexus, bin_time_of_arrival],
        params={
            InputFilepath: mcstas_file_path,
            MaximumProbability: DefaultMaximumProbability,
            TimeBinSteps: TimeBinSteps(50),
            EventWeightsConverter: event_weights_from_probability,
            ProtonChargeConverter: proton_charge_from_event_data,
        },
    )


def test_pipeline_builder(mcstas_workflow: sl.Pipeline, mcstas_file_path: str) -> None:
    from ess.nmx.mcstas_loader import InputFilepath

    assert mcstas_workflow.get(InputFilepath).compute() == mcstas_file_path


def test_pipeline_mcstas_loader(mcstas_workflow: sl.Pipeline) -> None:
    """Test if the loader graph is complete."""
    from ess.nmx.mcstas_loader import NMXData

    nmx_data = mcstas_workflow.compute(NMXData)
    assert isinstance(nmx_data, sc.DataGroup)
    assert nmx_data.sizes['panel'] == 3
    assert nmx_data.sizes['id'] == 1280 * 1280


def test_pipeline_mcstas_reduction(mcstas_workflow: sl.Pipeline) -> None:
    """Test if the loader graph is complete."""
    from ess.nmx.reduction import NMXReducedData

    nmx_reduced_data = mcstas_workflow.compute(NMXReducedData)
    assert isinstance(nmx_reduced_data, sc.DataGroup)
    assert nmx_reduced_data.sizes['panel'] == 3
    assert nmx_reduced_data.sizes['t'] == 50



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import Callable, Iterable, NewType, Optional

import scipp as sc
import scippnexus as snx

from .reduction import NMXData

PixelIDs = NewType("PixelIDs", sc.Variable)
InputFilepath = NewType("InputFilepath", str)
DetectorName = NewType("DetectorName", str)
DetectorBankName = NewType("DetectorBankName", str)

# McStas Configurations
MaximumProbability = NewType("MaximumProbability", int)
DefaultMaximumProbability = MaximumProbability(100_000)

McStasEventProbabilities = NewType("McStasEventProbabilities", sc.Variable)
EventWeights = NewType("EventWeights", sc.Variable)
EventWeightsConverter = NewType(
    "EventWeightsConverter",
    Callable[[MaximumProbability, McStasEventProbabilities], EventWeights],
)
"""A function that converts McStas probability to event weights."""

ProtonCharge = NewType("ProtonCharge", sc.Variable)
ProtonChargeConverter = NewType(
    "ProtonChargeConverter", Callable[[EventWeights], ProtonCharge]
)
"""A function that derives arbitrary proton charge based on event weights."""


def _retrieve_event_list_names(keys: Iterable[str]) -> tuple[str, ...]:
    import re

    mandatory_fields = 'p_x_y_n_id_t'
    # (weight, x, y, n, pixel id, time of arrival)
    pattern = r"bank(\d\d+)_events_dat_list_" + mandatory_fields

    def _filter_event_list_name(key: str) -> bool:
        return re.search(pattern, key) is not None

    if not (matching_keys := tuple(filter(_filter_event_list_name, keys))):
        raise ValueError("Can not find event list name.")

    return matching_keys


def _retrieve_raw_event_data(file: snx.File) -> sc.Variable:
    """Retrieve events from the nexus file."""
    bank_names = _retrieve_event_list_names(file["entry1/data"].keys())

    banks = [
        file["entry1/data/" + bank_name]["events"][()].rename_dims({'dim_0': 'event'})
        # ``dim_0``: event index, ``dim_1``: property index.
        for bank_name in bank_names
    ]

    return sc.concat(banks, 'event')


def _copy_partial_var(
    var: sc.Variable, idx: int, unit: Optional[str] = None, dtype: Optional[str] = None
) -> sc.Variable:
    """Retrieve a property from a variable."""
    var = var['dim_1', idx].astype(dtype or var.dtype, copy=True)
    if unit is not None:
        var.unit = sc.Unit(unit)
    return var


def _retrieve_crystal_rotation(file: snx.File, unit: str) -> sc.Variable:
    """Retrieve crystal rotation from the file."""

    return sc.vector(
        value=[file[f"entry1/simulation/Param/XtalPhi{key}"][...] for key in "XYZ"],
        unit=unit,
    )


def event_weights_from_probability(
    max_probability: MaximumProbability, probabilities: McStasEventProbabilities
) -> EventWeights:
    """Create event weights by scaling probability data.

    event_weights = max_probability * (probabilities / max(probabilities))

    Parameters
    ----------
    probabilities:
        The probabilities of the events.

    max_probability:
        The maximum probability to scale the weights.

    """
    maximum_probability = sc.scalar(max_probability, unit='counts')

    return EventWeights(maximum_probability * (probabilities / probabilities.max()))


def _compose_event_data_array(
    *,
    weights: sc.Variable,
    id_list: sc.Variable,
    t_list: sc.Variable,
    pixel_ids: sc.Variable,
    num_panels: int,
) -> sc.DataArray:
    """Combine data with coordinates loaded from the nexus file.

    Parameters
    ----------
    weights:
        The weights of the events.

    id_list:
        The pixel IDs of the events.

    t_list:
        The time of arrival of the events.

    pixel_ids:
        All possible pixel IDs of the detector.

    num_panels:
        The number of (detector) panels used in the experiment.

    """

    events = sc.DataArray(data=weights, coords={'t': t_list, 'id': id_list})
    grouped: sc.DataArray = events.group(pixel_ids)
    return grouped.fold(dim='id', sizes={'panel': num_panels, 'id': -1})


def proton_charge_from_event_data(event_da: sc.DataArray) -> ProtonCharge:
    """Make up the proton charge from the event data array.

    Proton charge is proportional to the number of neutrons,
    which is proportional to the number of events.
    The scale factor is manually chosen based on previous results
    to be convenient for data manipulation in the next steps.
    It is derived this way since
    the protons are not part of McStas simulation,
    and the number of neutrons is not included in the result.

    Parameters
    ----------
    event_da:
        The event data binned in detector panel and pixel id dimensions.

    """
    # Arbitrary number to scale the proton charge
    _proton_charge_scale_factor = sc.scalar(1 / 10_000, unit=None)

    return ProtonCharge(_proton_charge_scale_factor * event_da.bins.size().sum().data)


def load_mcstas_nexus(
    *,
    file_path: InputFilepath,
    event_weights_converter: EventWeightsConverter = event_weights_from_probability,
    proton_charge_converter: ProtonChargeConverter = proton_charge_from_event_data,
    max_probability: Optional[MaximumProbability] = None,
) -> NMXData:
    """Load McStas simulation result from h5(nexus) file.

    See :func:`~event_weights_from_probability` and
    :func:`~proton_charge_from_event_data` for details.

    Parameters
    ----------
    file_path:
        File name to load.

    event_weights_converter: :class:`~EventWeightsConverter`, \
        default: :func:`~event_weights_from_probability`
        A function to convert probabilities to event weights.
        The function should accept the probabilities as the first argument,
        and return the converted event weights.

    proton_charge_converter: :class:`~ProtonChargeConverter`, \
        default: :func:`~proton_charge_from_event_data`
        A function to convert the event weights to proton charge.
        The function should accept the event weights as the first argument,
        and return the proton charge.

    max_probability:
        The maximum probability to scale the weights.
        If not provided, ``DefaultMaximumProbability`` is used.

    """

    from .mcstas_xml import read_mcstas_geometry_xml

    geometry = read_mcstas_geometry_xml(file_path)
    detectors = [det.name for det in geometry.detectors]
    coords = geometry.to_coords(*detectors)

    with snx.File(file_path) as file:
        raw_data = _retrieve_raw_event_data(file)
        weights = event_weights_converter(
            max_probability or DefaultMaximumProbability,
            McStasEventProbabilities(
                _copy_partial_var(raw_data, idx=0, unit='counts')
            ),  # p
        )
        event_da = _compose_event_data_array(
            weights=weights,
            id_list=_copy_partial_var(raw_data, idx=4, dtype='int64'),  # id
            t_list=_copy_partial_var(raw_data, idx=5, unit='s'),  # t
            pixel_ids=coords.pop('pixel_id'),
            num_panels=len(detectors),
        )
        proton_charge = proton_charge_converter(event_da)
        crystal_rotation = _retrieve_crystal_rotation(
            file, geometry.simulation_settings.angle_unit
        )

    return NMXData(
        weights=event_da,
        proton_charge=proton_charge,
        crystal_rotation=crystal_rotation,
        **coords,
    )
